# This file is part of pymadcad,  distributed under license LGPL v3
'''
	implementation of views and displays for 3D
'''

import moderngl as mgl

from ...mathutils import uvec2, fvec2, fvec3, fmat2, fmat3, Box
from ..base import Scene



# class View:
#     ''' pure openGL 3D view over a madcad scene '''
#     scene: Scene
#     ''' madcad scene to render, it must contain 3D displays (at least a the root) '''
#     screen: mgl.FrameBuffer
#     ''' color map (and depth) openGL buffer '''
#     ident: mgl.FrameBuffer|None
#     ''' identification map (and depth) openGL buffer '''
#     targets: list
#     ''' exposed for `Scene` '''
#     uniforms: dict
#     ''' exposed for `Scene` '''
#     view: fmat3
#     ''' current view matrix, this will be the default for next rendering '''
#     def render(size=None, view=None)
# 
# class Offscreen(View):
#     color: ndarray
#     depth: ndarray|None
#     ident: ndarray|None
#     def render(size=None, view=None)
# 
# class QView(QWidget, View):
#     navigation: Pan|None
#     toolbar: QWidget|None
#     def control(key, event)
# 
# class Pan:
#     position: vec2
#     scale: float
#     orientation: complex
#     def adjust(box)
#     def center(vec3)
#     def zoom(ratio: float)
#     def pan(offset: vec2)
#     def rotate(offset: float)
#     def matrix() -> mat3

# class Projection(Display, View):
#     ''' 3d view in a 2d scene '''
#     world: mat3
#     size: vec2
#     limits: convex
# 
# class Frame(Display):
#     ''' rectangular frame carying few drawing specifications '''
# 



# This file is part of pymadcad,  distributed under license LGPL v3
from PIL import Image, ImageFont, ImageDraw
import numpy.core as np
import moderngl as mgl

from ..mathutils import fvec3, fvec4
from ..rendering import Display, highlight_color
from ..common import resourcedir
from .. import settings
from . import font_path, processalign


def create_font_texture(font: ImageFont, maxchar=1100) -> '(Image, (c,l))':
	''' create a font texture by rasterizing a given font '''
    # TODO: utiliser les methodes et attributs ImgeFont.size, .getmetrics(), etc pour avoir les hauteur et largeur de police
	
	# determine the size of the needed texture
	fontsize = font.size + 4
	width = 64
	while 2*(width//fontsize)**2 < maxchar:
		width *= 2
	c = width//(fontsize//2)
	l = maxchar//c +1
	width = c * (fontsize//2)
	height = l * fontsize
	# create texture
	tex = Image.new('L', (width,height), 0)
	draw = ImageDraw.Draw(tex)
	# draw the font onto
	for i in range(maxchar):
		draw.text(char_placement(font.size, c, l, i), chr(i), fill=255, font=font)
	#tex.show()
	
	return tex, (c, l)

def char_placement(fontsize, c, l, n):
	fontsize += 4
	return ((fontsize//2) * (n%c), fontsize * (n//c))


class TextDisplay(Display):
	''' halo display of a monospaced text 
	
		This class is usually used through `scheme.note_floating()`
	'''
	pointsdef = [
		[0, 0],
		[0,-1],
		[1,-1],
		[0, 0],
		[1,-1],
		[1, 0],
		]
	
	def __init__(self, scene, position, text, size=None, color=None, font=None, align=(0,0), layer=0):
		if not color:	color = settings.display['annotation_color']
		if not size:	size = settings.display['view_font_size']
		self.position = fvec3(position)
		self.color = fvec3(color)
		self.size = size
		self.layer = layer
		self.selected = False
		self.hovered = False
		
		# load font
		fontfile = font_path(font or 'NotoMono-Regular')
		def load(scene):
			img, align = create_font_texture(ImageFont.truetype(fontfile, 2*size))
			return scene.context.texture(img.size, 1, img.tobytes()), align
		self.fonttex, self.fontalign = scene.share(('fontfile', size), load)
		
		# load shader
		def load(scene):
			shader = scene.context.program(
						vertex_shader=open(resourcedir+'/shaders/font.vert').read(),
						fragment_shader=open(resourcedir+'/shaders/font.frag').read(),
						)
			shader['fonttex'].value = 0
			return shader
		self.shader = scene.share('shader_font', load)
		
		# place triangles
		points = np.zeros((len(self.pointsdef)*len(text), 4), 'f4')
		l = 0
		c = 0
		mc = 0
		for i,char in enumerate(text):
			if char == '\n':
				c = 0
				l += 1
			elif char == '\t':
				c += 4 - c%4	# TODO: use a tab size in settings
			elif char == ' ':
				c += 1
			else:
				n = ord(char)
				#placement = char_placement(2*size, *self.fontalign, n)
				for j,add in enumerate(self.pointsdef):
					points[len(self.pointsdef)*i + j] = [
						add[0]+c, 
						add[1]-l, 
						(n % self.fontalign[0] +add[0]) / self.fontalign[0], 
						(n //self.fontalign[0] -add[1]) / self.fontalign[1],
						]
				c += 1
			if c > mc:
				mc = c
		l += 1		
		align = (processalign(align[0], mc), -processalign(align[1], l))
		points -= [*align, 0, 0]
		self.textsize = (mc,l)
		self.visualsize = (-align[0], -align[1], mc//2-align[0], l-align[1])
			
		
		# create the buffers on the GPU
		self.vb_points = scene.context.buffer(points)
		self.vb_faces = scene.context.buffer(np.arange(points.shape[0], dtype='u4'))
		self.va = scene.context.vertex_array(
			self.shader,
			[(self.vb_points, '2f 2f', 'v_position', 'v_uv')],
			)

	def stack(self, view):
		return (self, 'screen', 2, self._render),
	
	def _render(self, view):		
		self.shader['layer'] = self.layer
		self.shader['color'].write(highlight_color(self, self.color))
		self.shader['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self.shader['view'].write(view.uniforms['view'])
		self.shader['proj'].write(view.uniforms['proj'])
		self.shader['ratio'].value = (
				(self.size-2.5) / view.uniforms['size'].x*2,
				(self.size-2.5) / view.uniforms['size'].y*4,
				)
		self.fonttex.use(0)
		self.va.render(mgl.TRIANGLES)



from __future__ import annotations

import moderngl as mgl
from numpy import ndarray
import numpy as np

from .. import settings
from ..mathutils import ivec2, uvec2, fvec4, mix


def writeproperty(func):
	''' Decorator to create a property that has only an action on variable write '''
	fieldname = '_'+func.__name__
	def getter(self):	
		return getattr(self, fieldname)
	def setter(self, value):
		setattr(self, fieldname, value)
		func(self)
	return property(getter, setter, doc=func.__doc__)
	
def forwardproperty(attribute, sub):
	''' read/write property to a attribute's attribute '''
	def getter(self):
		return getattr(getattr(self, attribute), sub)
	def setter(self, value):
		setattr(getattr(self, attribute), sub, value)
	return property(getter, setter)

def sceneshare(generator, immortal=False):
	''' decorator for shared ressources with the scene '''
	name = generator
	def share(self, scene):
		if not scene.shared.get(name):
			data = generator(self, scene)
			if immortal:
				data = Rc(data)
			else:
				data = Weak(data)
			scene.shared[name] = data
		setattr(self, generator.__name__, Rc(scene.shared[name]))
		vars(self).update(scene.shared[name]())
	share.__doc__ = generator.__doc__
	return share

def receiver(generator) -> Callable:
	''' simple helper running the generator until the first yield, and returning its send method '''
	next(generator, None)
	def callback(evt):
		try:
			generator.send(evt)
		except StopIteration:
			return False
		else:
			return True
	return callback
	
class Weak:
	''' weak reference to an object '''
	__slots__ = 'object', 'refcount'
	def __init__(self, object):
		self.object = object
		self.refcount = 0
	def __bool__(self):
		''' return True if the object is still alive and has not been dropped '''
		return self.refcount > 0
	def __call__(self):
		''' retreive the underlying object or None if deallocated '''
		return self.object

class Rc:
	''' strong counted reference to an object '''
	__slots__ = 'weak'
	def __init__(self, other):
		if isinstance(other, Weak):
			self.weak = other
		elif isisntance(other, Rc):
			self.weak = other.weak
		else:
			self.weak = Weak(other)
		self.weak.refcount += 1
	def __del__(self):
		if self.weak is not None:
			self.weak.refcount -= 1
			if self.weak.refcount <= 0:
				self.weak.object = None
		self.weak = None
	def __call__(self):
		''' retreive the underlying object '''
		return self.weak.object

class CheapMap:
	''' object retreiving portions of a framebuffer, on demand and avoiding redundant memory copies '''
	def __init__(self, framebuffer: mgl.Framebuffer, attachment:int):
		self.framebuffer = framebuffer
		self.attachment = attachment
		if attachment < 0:
			renderbuffer = framebuffer.depth_attachment
		else:
			renderbuffer = framebuffer.color_attachments[attachment]
		self.dtype = renderbuffer.dtype
		self.components = renderbuffer.components
		size = renderbuffer.size
		self.buffer = np.zeros(size[1]*size[0], self.dtype)
		
		self.clear()
	
	def clear(self):
		''' clear the last retreived region
			
			next call to `region` will trigger a transfer from the GPU 
		'''
		self.viewport = (1, 1, -1, -1)
		
	def region(self, viewport: tuple) -> ndarray:
		''' retreive a given region of the framebuffer from the GPU
		
			any next extracted region that is inside this one will no need a new transfer
		'''
		if (viewport[0] < self.viewport[0] or viewport[2] > self.viewport[2] 
		or  viewport[1] < self.viewport[1] or viewport[3] > self.viewport[3]):
			self.viewport = viewport
			self.framebuffer.read_into(self.buffer, 
				(viewport[0], self.framebuffer.height-viewport[3], viewport[2]-viewport[0], viewport[3]-viewport[1]), 
				attachment=self.attachment, components=self.components, dtype=self.dtype)
		
		view = self.buffer[:(self.viewport[3]-self.viewport[1]) * (self.viewport[2]-self.viewport[0])]
		view = view.reshape((self.viewport[3]-self.viewport[1], self.viewport[2]-self.viewport[0]))
		view = view[::-1]
		return view[
			viewport[1]-self.viewport[1]:viewport[3]-self.viewport[1],
			viewport[0]-self.viewport[0]:viewport[2]-self.viewport[0],
			]


def snail(radius):
	''' Generator of coordinates snailing around 0,0 '''
	x = 0
	y = 0
	for r in range(radius):
		for x in range(-r,r):		yield ivec2(x,-r)
		for y in range(-r,r):		yield ivec2(r, y)
		for x in reversed(range(-r,r)):	yield ivec2(x, r)
		for y in reversed(range(-r,r)):	yield ivec2(-r,y)

def snailaround(pt, box, radius):
	''' Generator of coordinates snailing around pt, coordinates that goes out of the box are skipped '''
	cx,cy = pt
	mx,my = box
	for rx,ry in snail(radius):
		x,y = cx+rx, cy+ry
		if 0 <= x and x < mx and 0 <= y and y < my:
			yield ivec2(x,y)

def glsize(size: uvec2) -> uvec2:
	# if the size is not a multiple of 4, it seems that openGL or Qt doesn't understand its strides well
	m = 4
	return size + uvec2(-ivec2(size)%m)

def highlight_color(display: Display, color: fvec3) -> fvec3:
	if display.selected:    highlight = fvec4(settings.display['selection_color'])
	elif display.hovered:   highlight = fvec4(settings.display['hover_color'])
	else:                   highlight = fvec4(0)
	return mix(color, highlight.rgb, highlight.a)
			
# qt conversion functions

try:
	from ..qt import QPoint, QSize
except ImportError:
	pass
else:
	def vec_to_qpoint(p: ivec2) -> QPoint: 	return QPoint(p.x, p.y)
	def qpoint_to_vec(p: QPoint) -> ivec2:	return ivec2(p.x(), p.y())
	def vec_to_qsize(p: uvec2) -> QSize: 	return QSize(p.x, p.y)
	def qsize_to_vec(p: QSize) -> uvec2:	return uvec2(p.width(), p.height())



# This file is part of pymadcad,  distributed under license LGPL v3
from __future__ import annotations
import traceback
from weakref import WeakValueDictionary
from dataclasses import dataclass
from copy import deepcopy
from operator import attrgetter

import moderngl as mgl

from .. import settings
from ..mathutils import boundingbox
from .utils import writeproperty

__all__ = ['Scene', 'Step', 'Display', 'Group', 'Displayable']


# minimum opengl version required by the rendering pipeline
opengl_version = (4,3)
# shared open gl context, None if not yet initialized
global_context = None
# helper constant to avoid reinstantiating empty tuples
empty = ()



class Scene:
	''' rendering pipeline for madcad displayable objects 
	
		This class is gui-agnostic, it only relies on OpenGL, and the context has to be created by the user.
		
		When an object is added to the scene, a Display is not immediately created for it, the object is put into the queue and the Display is created at the next render.
		If the object is removed from the scene before the next render, it is dequeued.
	'''
	root: Display
	''' the root display of the scene, usually a `Group` '''
	stacks: dict[str, list[Frame]]
	''' a list of the display callbacks for each target frame buffers '''
	touched: bool
	''' trigger for restacking render steps '''
	overrides: dict[type, Callable] = {}
	''' attribute allowing to alter how the displayable are transformed into displays
	
		- the class attribute stores default overrides
		- the instance attribute stores the effective overrides for the scene
	'''
	options: dict
	''' scene rendering parameters, deriving from `settings.scene` '''
	shared: dict
	''' strong references to the resources we don't want to free when no display is using them '''
	context: mgl.Context
	''' opengl context used in this scene's displays. It is assumed to have `gc_mode = auto` '''
	selection: set[Display]
	''' current set of selected displays (readonly, use the scene methods to alter the selection) '''
	hover: Display|None = None
	''' current display being hovered (readwrite) '''
	
	def __init__(self, src:object=None, options:dict=None, context:mgl.Context=None, overrides:dict=None):
		'''
		Args:
			src:        the root object of the scene, it must be a displayable and is usually a `dict`
			options:    scene rendering parameters, overriding `settings.scene`
			context:    the openGL context to use to send rendering instructions to the GPU
		'''
		self.root = None
		self.overrides = deepcopy(Scene.overrides)
		if overrides:
			self.overrides.update(overrides)
		self.stacks = {}
		self.options = deepcopy(settings.scene)
		self.shared = {}
		self.selection = set()
		self.touched = False
		self._hover = None
		self._memo = set()
		if options:   self.options.update(options)
		
		global global_context
		if context:
			self.context = context
		elif global_context:
			self.context = global_context
		else:
			self.context = global_context = mgl.create_standalone_context(requires=opengl_version)
			self.context.gc_mode = 'auto'
			self.context.line_width = settings.display["line_width"]
		
		self.update(src)
	
	def update(self, src:object):
		''' update the root display of the scene '''
		if src is None:
			src = {}
		self.root = self.display(src, self.root)
		self.root.key = ()
		self.touch()
		
	def touch(self):
		''' inform the scene than its composition changed and that it needs to recompute its call stack '''
		self.touched = True
		
	def display(self, obj, former:Display=None) -> Display:
		''' Create a display for the given object for the current scene.
		
			This is the actual function converting objects into displays.
			You don't need to call this method if you just want to add an object to the scene, use add() instead
		'''
		# prevent reference-loop in groups (groups are taken from the execution env, so the user may not want to display it however we are trying to)
		ido = id(obj)
		assert ido not in self._memo, 'there should not be recursion loops in cascading displays'
		self._memo.add(ido)
		
		try:
			# no refresh if object has not changed
			if hasattr(former, 'source') and former.source == obj:
				disp = former
			# refresh cleverly if a previous displays proposes it
			elif former and former.update(self, obj):
				disp = former
			# create it with overrides
			elif type(obj) in self.overrides:
				disp = self.overrides[type(obj)](self, obj)
			# create it with the displayable protocol
			elif hasattr(obj, 'display'):
				if isinstance(obj, type):
					raise TypeError('types are not displayable')
				if isinstance(obj.display, type):
					disp = obj.display(self, obj)
				elif callable(obj.display):
					disp = obj.display(self)
				else:
					raise TypeError("member 'display' must be a method or a type, on {}".format(type(obj).__name__))
			else:
				raise TypeError('type {} is not displayable'.format(type(obj).__name__))
			
			if not isinstance(disp, Display):
				raise TypeError('the display for {} is not a subclass of Display: {}'.format(type(obj).__name__, type(disp)))
				
		finally:
			self._memo.remove(ido)
		
		# keeping a reference of the source object may increas the RAM used but avoid to refresh displays when updating the scene with the same constant value
		if self.options['track_source']:
			disp.source = obj
		return disp
		
	def displayable(self, obj) -> bool:
		''' return whether the given object could be displayed in the scene '''
		if type(obj) in self.overrides:
			return True
		elif hasattr(obj, 'display'):
			if isinstance(obj, type):
				return False
			if isinstance(obj.display, type):
				return True
			elif callable(obj.display):
				return True
		return False
	
	def share(self, key, generator=None):
		''' Get a shared resource loaded or load it using the function func.
			If func is not provided, an error is raised
		'''
		if key in self.shared:	
			return self.shared[key]
		elif callable(generator):
			res = generator(self)
			self.shared[key] = res
			return res
		else:
			raise KeyError("resource {} doesn't exist or is not loaded".format(repr(key)))
	
	def prepare(self):
		''' convert all displayable to displays and bufferize everything for comming renderings '''
		if self.touched:
			self.touched = False
			self.active = None
			self.hover = None
			with self.context:
				for stack in self.stacks.values():
					stack.clear()
				for step in self.root.stack(self):
					if isinstance(step, tuple):
						step = Step(*step)
					elif not isinstance(step, Step):
						raise TypeError('stack step must be of type Step, not {}'.format(type(step)))
					if not isinstance(step.display, Display):
						raise TypeError('step.display must be a Display, not {}'.format(type(step.display)))
					if step.target not in self.stacks:	
						self.stacks[step.target] = []
					stack = self.stacks[step.target]
					stack.append(step)
				# sort the stack using the specified priorities
				for stack in self.stacks.values():
					stack.sort(key=attrgetter('priority'))
	
	def render(self, view):
		''' Render to the view targets. 
			
			This must be called by the view widget, once the OpenGL context is set.
		'''
		with self.context:
			# restack and bufferize
			try:
				self.prepare()
			except Exception:
				traceback.print_exc()
			# render everything
			for target, frame, setup in view.targets:
				try:
					view.target = frame
					frame.use()
					setup()
					for step in self.stacks.get(target,empty):
						step.render(view)
				except Exception:
					traceback.print_exc()
	
	def selection_add(self, display:Display, sub:int=None):
		''' select the given display '''
		if not hasattr(display, 'selected'):
			raise TypeError('{} is not selectable'.format(type(display).__name__))
		if isinstance(display.selected, set):
			display.selected.add(sub)
			display.selected = display.selected
			self.selection.add(display)
		else:
			display.selected = True
			self.selection.add(display)
		self.touch()
	
	def selection_remove(self, display:Display, sub:int=None):
		''' deselect the given display '''
		if not hasattr(display, 'selected'):
			raise TypeError('{} is not selectable'.format(type(display).__name__))
		if isinstance(display.selected, set):
			if sub is None:
				display.selected.clear()
			else:
				display.selected.discard(sub)
			display.selected = display.selected
			if not display.selected:
				self.selection.discard(display)
		else:
			display.selected = False
			self.selection.discard(display)
		self.touch()
		
	def selection_toggle(self, display:Display, sub:int=None):
		if not hasattr(display, 'selected'):
			raise TypeError('{} is not selectable'.format(type(display).__name__))
		if isinstance(display.selected, set):
			selected = sub in display.selected
		else:
			selected = display.selected
		if selected:
			self.selection_remove(display, sub)
		else:
			self.selection_add(display, sub)
	
	def selection_clear(self):
		''' deselect all previously selected displays '''
		for display in self.selection:
			if isinstance(display.selected, set):
				display.selected.clear()
				display.selected = display.selected
			else:
				display.selected = False
		if self.selection:
			self.touch()
			self.selection.clear()
	
	def _get_hover(self):
		return self._hover
	def _set_hover(self, hover):
		# hover can be set to a tuple to hover a subitem
		if isinstance(hover, tuple):
			hover, sub = hover
		else:
			sub = None
		# prevent unnecessary hover change to avoid triggering rebuffering of hover settings
		if hover is self._hover:
			if hover and isinstance(hover.hovered, set):
				if sub in hover.hovered:
					return
			else:
				return
		# clear previous hover in displays
		if self._hover:
			if isinstance(self._hover.hovered, set):
				self._hover.hovered.clear()
				self._hover.hovered = self._hover.hovered
			else:
				self._hover.hovered = False
		self._hover = hover
		# set new hover in displays
		if self._hover:
			if isinstance(self._hover.hovered, set):
				self._hover.hovered.add(sub)
				self._hover.hovered = self._hover.hovered
			else:
				self._hover.hovered = True
	hover = property(_get_hover, _set_hover)

	def __getitem__(self, key):
		return self.root[key]
	
	def __setitem__(self, key, value):
		self.root[key] = value
	
	def item(self, key:tuple):
		node = self.root
		for sub in key:
			node = node[sub]
	
@dataclass
class Step:
	''' describes a rendering step for a display '''
	display: Display
	''' the display instance responsible of this callback '''
	target: str
	''' the name of the render target in the view that will be rendered (see View) '''
	priority: float
	''' a float that is used to insert the callable at the proper place in the rendering stack '''
	render: Callable
	''' a function that renders, signature is  `func(view)` '''


class Display:
	''' base class for objects displayed in a scene '''
	key: str = None
	''' display path in the scene tree, set when Scene.restack is called '''
	source: object
	''' source object of that display, set by `Scene.display` if this option is enabled '''
	world: fmat3|fmat4
	''' (optional) matrix from local space to scene space '''
	box: Box
	''' (optional) boudingbox of the display in local space '''
	selected: bool|set
	''' (optional) whether the display has been selected by the user 
	
		when display sets this value as a set, it designate selection of internal parts of the display designed by integers (like groups in a mesh). when the whole display is selected and not a subpart of it, `None` is present in the set.
		This doesn't prevent the scene from trying to assign booleans to it so it must be protected by a setter adding `None` to the set instead of assigning to `True`
	'''
	hovered: bool|set
	''' (optional) whether the display is hovered by a pointing device and might be selected 
	
		the same semantics of `selected` applies here
	'''
	
	def __init__(self, *args): 
		pass
	
	def display(self, scene:Scene) -> Self:
		''' Displays are obviously displayable as themselves, this should be overriden '''
		return self
	
	def update(self, scene:Scene, src:object) -> bool:
		''' Update the current displays internal datas with the given displayable .
			
			If the display cannot be upgraded, it must return False to be replaced by a fresh new display created from the displayable
		'''
		return False
	
	def stack(self, scene:Scene) -> Iterable[Step]:
		''' Rendering functions to insert in the renderpipeline.
		
			The expected result can be any iterable providing tuples `(key, target, priority, callable)` such as:
			
			The view contains the uniforms, rendering targets and the scene for common resources
		'''
		return empty
	
	def control(self, view, key:tuple, sub:tuple, evt:QEvent):
		''' Handle input events occuring on the area of this display (or of one of its subdisplay).
			For subdisplay events, the parents control functions are called first, and the sub display controls are called only if the event is not accepted by parents
			
			Parameters:
				key:    the key path for the current display
				sub:    the key path for the subdisplay
				evt:    the Qt event (see Qt doc)
		'''
		pass
	
	def __iter__(self) -> iter:
		''' yield child displays '''
		return iter(empty)
	
	def __getitem__(self, key) -> Display:
		''' Get a child display by its key in this display (like in a scene) '''
		raise IndexError('{} has no sub displays'.format(type(self).__name__))


class Group(Display):
	''' display holding multiple displays associated with a key, just like a dictionnary 
	
		`world`, `selected`, `hovered` properties are propagated to its children
	'''
	displays: dict
	''' dictionnary of displays currently displayed in the group '''
	
	def __init__(self, scene:Scene, src:dict|list=None, world=1):
		self.displays = {}
		self._pending = None
		self._world = world
		self._selected = False
		self._hovered = False
		if src:
			self.update(None, src)
		
	@property
	def box(self):
		return boundingbox(
			(getattr(display, 'box', None)   for display in self.displays.values()), 
			ignore=True)
	
	@writeproperty
	def world(self):
		for display in self.displays.values():
			display.world = self.world
			
	@writeproperty
	def selected(self):
		for display in self.displays.values():
			if not hasattr(display, 'selected'):
				continue
			display.selected = self.selected
	
	@writeproperty
	def hovered(self):
		for display in self.displays.values():
			if not hasattr(display, 'hovered'):
				continue
			display.hovered = self.hovered
	
	def update(self, *args):
		''' 
			update(scene, src)
			update(src)
		
			update all child displays with the displayable present in the given dictionnary
			
			former children that do not match keys in the given dictionnary are dropped
			
			Note:
				the new content will not be immediately available in `self.displays` because they will only be buffered at next rendering. it makes this function thread safe and able to run without a reference to the scene
		'''
		if len(args) == 1:
			src, = args
		elif len(args) == 2:
			scene, src = args
		else:
			raise TypeError("expected 1 or 2 arguments, got {}".format(len(args)))
		if isinstance(src, list):
			src = dict(enumerate(src))
		if not isinstance(src, dict):
			return False
		self._pending = src
		return True
	
	def prepare(self, scene):
		# update with the pending data
		if self._pending is not None:
			removed = []
			for key in self.displays:
				if key not in self._pending or not scene.displayable(self._pending[key]):
					removed.append(key)
			for key in removed:
				del self.displays[key]
			for key, displayable in self._pending.items():
				if not scene.displayable(displayable):
					continue
				try:
					self.displays[key] = scene.display(displayable, self.displays.get(key))
				except:
					print('\ntried to display', object.__repr__(displayable))
					traceback.print_exc()
					continue
			self._pending = None
	
	def stack(self, scene):
		self.prepare(scene)
		for key, display in self.displays.items():
			display.world = self.world
			display.key = (*self.key, key)
			yield from display.stack(scene)
	
	def __iter__(self):
		return iter(self.displays.values())
	
	def __getitem__(self, key):
		return self.displays[key]

	def __setitem__(self, key, value):
		if self.pending is None:
			self.pending = copy(self.displays)
		self.pending[key] = value

Scene.overrides[dict] = Group
Scene.overrides[list] = Group
Scene.overrides[type(None)] = Display


class Displayable:
	''' Simple displayable initializing the given Display class with arguments 
		
		At the display creation time, it will simply execute `build(*args, **kwargs)`
	'''
	__slots__ = 'build', 'args', 'kwargs'
	def __init__(self, build, *args, **kwargs):
		self.args, self.kwargs = args, kwargs
		self.build = build
	def __repr__(self):
		return '{}({}, {})'.format(
			type(self).__name__, 
			','.join(repr(arg) for arg in self.args), 
			','.join(key+'='+repr(arg)  for key, arg in self.kwargs.items())
			)
	def display(self, scene):
		return self.build(scene, *self.args, **self.kwargs)



import moderngl as mgl

from ...mathutils import *
from ...common import resourcedir
from ...mesh import typedlist_to_numpy
from .. import Display


class VoxelsDisplay(Display):
	''' display a voxel as a scalar field.
		its opacity is computed using its density as absorbance.
		
		Parameters:
			
			voxel:	a numpy array with dimension 3, no specific value range is required
			
		Attributes:
		
			space:	transformation matrix from voxel normalized coordinates (position in the array as range 0-1) to local coordinates
			world:	transformation matrix from local to world coordinates
			value_range:	
				
				tuple `(min, max)`  giving the min and max values to display
				- values below min will be transparent
				- values above max will be fully opaq
				
			color_range:	tuple `(fvec4, fvec4)` of colors matching the value range bounds
		
		value examples:
		
			* value_range
			
				```
				(0.4, 0.9)  # erase the lower value and saturate the maxium ones
				```
				
			* color_range
			
				```
				(fvec4(0,0,1,1), fvec4(0,1,0,1))	 # blue-green
				(fvec4(0,0.5,1,1), fvec4(1,1,0,1))   # lightblue-yellow
				(fvec4(1,0,0.3,1), fvec4(1,1,0,1))   # red-yellow
				(fvec4(0,0,1,1), fvec4(1,0.5,0.1,1)) # blue-orange
				(fvec4(0,0,1,1), fvec4(1,0.1,0.1,1)) # blue-red
				(fvec4(0,0,0,1), fvec4(1,1,1,1))     # black-white
				(fvec4(1,1,1,1), fvec4(1,1,1,1))     # white-white
				```
	'''
	def __init__(self, scene, voxel: 'ndarray', space: 'fmat4', 
				value=(0, 1), 
				color=(fvec4(0,0,1,1), fvec4(0,1,0,1))):
		
		self.voxel = scene.context.texture3d(
			voxel.shape, 
			1, 
			voxel.transpose(2,1,0).copy().astype('f2'),  # axis permutation for opengl
			dtype='f2',
			)
		self.voxel.repeat_x = False
		self.voxel.repeat_y = False
		self.voxel.repeat_z = False
		self.space = space
		self.value_range = value
		self.color_range = color
		
		self.shader, self.vb = scene.share(type(self), self._share)
		self.va = scene.context.vertex_array(
				self.shader,
				[(self.vb, '3f', 'v_position')],
				)
	
	def _share(self, scene):
		from ... import generation
		
		# load shader
		shader = scene.context.program(
					vertex_shader=open(resourcedir+'/shaders/voxel.vert').read(),
					fragment_shader=open(resourcedir+'/shaders/voxel.frag').read(),
					)
		# load vertex buffer for the brick
		brick = generation.brick(min=vec3(0), max=vec3(1)) .flip()
		pts = typedlist(vec3)
		for face in brick.faces:
			pts.extend(brick.facepoints(face))
		vb = scene.context.buffer(typedlist_to_numpy(pts, 'f4'))
		
		return shader, vb
	
	def stack(self, scene):
		return (self, 'screen', 10, self._render),

	def _render(self, view):
		view.scene.context.enable_only(mgl.DEPTH_TEST | mgl.BLEND | mgl.CULL_FACE)
		self.voxel.use(0)
		self.va.program['value_min'] = self.value_range[0]
		self.va.program['value_max'] = self.value_range[1]
		self.va.program['color_min'].write(self.color_range[0])
		self.va.program['color_max'].write(self.color_range[1])
		self.va.program['view'].write(view.uniforms['view'] * self.world * self.space)
		self.va.program['proj'].write(view.uniforms['proj'])
		self.va.render(mgl.TRIANGLES)
		view.scene.context.enable_only(mgl.DEPTH_TEST | mgl.BLEND)



# This file is part of pymadcad,  distributed under license LGPL v3
import numpy.core as np
import moderngl as mgl
from PIL import Image

from ... import settings, primitives
from ...mathutils import *
from ...mesh import typedlist_to_numpy
from ...common import resourcedir
from ...qt import Qt, QEvent
from .. import Display, Scene, writeproperty, highlight_color
from . import npboundingbox, load_shader_ident, load_shader_subident, load_shader_wire, load_shader_uniformcolor

__all__ = ['PointDisplay', 'AxisDisplay', 'BoxDisplay', 'SplineDisplay', 'GridDisplay']


class PointDisplay(Display):
	position: fvec3
	size: float
	color: fvec3
	selected: bool
	hovered: bool
			
	@property
	def box(self):
		return Box(center=self.position, width=fvec3(0))
	
	def __init__(self, scene, position, size=10, color=None):
		self.position = fvec3(position)
		self.size = size
		self.color = fvec3(color or settings.colors['line'])
		self.selected = False
		self.hovered = False

		vars(self).update(scene.share(type(self), self._shared))
	
	def _shared(self, scene):
		img = Image.open(resourcedir+'/textures/point.png')
		texture = scene.context.texture(img.size, 1, img.convert('L').tobytes())
		shader = scene.context.program(
								vertex_shader=open(resourcedir+'/shaders/pointhalo.vert').read(),
								fragment_shader=open(resourcedir+'/shaders/pointhalo.frag').read(),
								)
		shader['halotex'].value = 0
		ident_shader = scene.context.program(
								vertex_shader=open(resourcedir+'/shaders/pointhalo-ident.vert').read(),
								fragment_shader=open(resourcedir+'/shaders/ident.frag').read(),
								)
		vb = scene.context.buffer(np.array([(0,0), (0,1), (1,1), (0,0), (1,1), (1,0)], 'f4'))
		
		return dict(
				_texture = texture,
				_va_screen = scene.context.vertex_array(shader, [(vb, '2f', 'v_uv')], mode=mgl.TRIANGLES),
				_va_ident = scene.context.vertex_array(ident_shader, [(vb, '2f', 'v_uv')], mode=mgl.TRIANGLES),
				)
	
	def stack(self, scene):
		return ( (self, 'ident', 2, self._identify),
				(self, 'screen', 2, self._render))

	def _render(self, view):
		self._va_screen.program['color'].write(highlight_color(self, self.color))
		self._va_screen.program['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self._va_screen.program['view'].write(view.uniforms['view'])
		self._va_screen.program['proj'].write(view.uniforms['proj'])
		self._va_screen.program['ratio'] = (
				self.size / view.screen.width,
				self.size / view.screen.height,
				)
		self._texture.use(0)
		self._va_screen.render()

	def _identify(self, view):
		self._va_ident.program['ident'] = view.identstep(1)
		self._va_ident.program['position'].write(fvec3(self.world * fvec4(self.position,1)))
		self._va_ident.program['view'].write(view.uniforms['view'])
		self._va_ident.program['proj'].write(view.uniforms['proj'])
		self._va_ident.program['ratio'] = (
				1.5 * self.size / view.screen.width,
				1.5 * self.size / view.screen.height,
				)
		self._va_ident.render()


class AxisDisplay(Display):
	origin: fvec3
	direction: fvec3
	interval: (float, float)
	color: fvec3
	selected: bool
	hovered: bool

	pattern = [0, 0.25, 0.45, 0.55, 0.75, 1]
	repetitions = 3
	
	@property
	def box(self):
		return Box(center=self.origin, width=fvec3(0))
	
	def __init__(self, scene, axis, interval=None, color=None):
		self.origin = fvec3(axis[0])
		self.direction = fvec3(axis[1])
		self.interval = interval
		self.color = fvec3(color or settings.colors['line'])
		self.selected = False
		self.hovered = False
		
		vars(self).update(scene.share(type(self), self._shared))
	
	def _shared(self, scene):
		shader = scene.context.program(
					vertex_shader=open(resourcedir+'/shaders/axis.vert').read(),
					fragment_shader=open(resourcedir+'/shaders/axis.frag').read(),
					)
		ident_shader = scene.context.program(
					vertex_shader=open(resourcedir+'/shaders/axis-ident.vert').read(),
					fragment_shader=open(resourcedir+'/shaders/ident.frag').read(),
					)
		pts = []
		for i in range(-1, self.repetitions+1):
			for pt in self.pattern:
				if i == -1:					alpha = pt
				elif i == self.repetitions:	alpha = 1-pt
				else:						alpha = 1
				pts.append(((pt+i)/self.repetitions, alpha))
		vb = scene.context.buffer(np.array(pts, 'f4'))
		return dict(
			_va_screen = scene.context.vertex_array(shader, [(vb, 'f f', 'v_absciss', 'v_alpha')], mode=mgl.LINES),
			_va_ident = scene.context.vertex_array(ident_shader, [(vb, 'f 4x', 'v_absciss')], mode=mgl.LINES),
			)
	
	def stack(self, scene):
		return ( (self, 'ident', 2, self._identify),
				(self, 'screen', 2, self._render))
	
	def _render(self, view):
		self._va_screen.program['color'].write(highlight_color(self, self.color))
		self._va_screen.program['projview'].write(view.uniforms['projview'])
		self._va_screen.program['world'].write(self.world)
		self._va_screen.program['origin'].write(self.origin)
		self._va_screen.program['direction'].write(self.direction)
		self._va_screen.program['interval'] = self._disp_interval(view)
		self._va_screen.render()
	
	def _identify(self, view):
		self._va_ident.program['projview'].write(view.uniforms['projview'])
		self._va_ident.program['world'].write(self.world)
		self._va_ident.program['origin'].write(self.origin)
		self._va_ident.program['direction'].write(self.direction)
		self._va_ident.program['interval'] = self._disp_interval(view)
		self._va_ident.program['ident'] = view.identstep(1)
		self._va_ident.render()
	
	def _disp_interval(self, view):
		if self.interval:	
			return self.interval
		else:
			size = -view.uniforms['view'][3][2]/6
			return (-0.5*size, size)
	


class AnnotationDisplay(Display):
	box: Box
	color: fvec4
	selected: bool
	hovered: bool
	
	def __init__(self, scene, points, color):
		self.color = fvec3(color or settings.colors['annotation'])
		self.box = npboundingbox(points)
		self.selected = False
		self.hovered = False
		
		shader_screen = scene.share(AnnotationDisplay, self._shared)
		shader_ident = scene.share(load_shader_ident, load_shader_ident)
		
		# allocate buffers
		vb = scene.context.buffer(points)
		self._va_screen = scene.context.vertex_array(shader_screen, [(vb, '3f f', 'v_position', 'v_alpha')], mode=mgl.LINES)
		self._va_ident = scene.context.vertex_array(shader_ident, [(vb, '3f 4x', 'v_position')], mode=mgl.LINES)
	
	def _shared(self, scene):
		return scene.context.program(
			vertex_shader=open(resourcedir+'/shaders/annotation.vert').read(),
			fragment_shader=open(resourcedir+'/shaders/annotation.frag').read(),
			)
	
	def stack(self, scene):
		return ( (self, 'ident', 2, self._identify),
				(self, 'screen', 2, self._render)) 
	
	def _render(self, view):
		self._va_screen.program['color'].write(highlight_color(self, self.color))
		self._va_screen.program['proj'].write(view.uniforms['proj'])
		self._va_screen.program['view'].write(view.uniforms['view'] * self.world)
		self._va_screen.render()
	
	def _identify(self, view):
		self._va_ident.program['ident'] = view.identstep(1)
		self._va_ident.program['proj'].write(view.uniforms['proj'])
		self._va_ident.program['view'].write(view.uniforms['view'] * self.world)
		self._va_ident.render()

class BoxDisplay(AnnotationDisplay):
	def __init__(self, scene, box, color=None):
		# place points
		x,y,z = box.width
		c = 0.1*min(x,y,z)	# corner
		o = 0.4 # wire alpha
		pts = np.array([
			# corner
			(0,0,0,1),(c,0,0,1),
			(0,0,0,1),(0,c,0,1),
			(0,0,0,1),(0,0,2*c,1),
			
			# box wire
			(0,0,0,o),(x,0,0,o),
			(0,0,0,o),(0,y,0,o),
			(0,0,0,o),(0,0,z,o),
			
			(x,0,0,o),(x,y,0,o),
			(x,0,0,o),(x,0,z,o),
			(0,y,0,o),(x,y,0,o),
			(0,y,0,o),(0,y,z,o),
			(0,0,z,o),(x,0,z,o),
			(0,0,z,o),(0,y,z,o),
			
			(x,y,0,o),(x,y,z,o),
			(x,0,z,o),(x,y,z,o),
			(0,y,z,o),(x,y,z,o),
			], dtype='f4')
		pts += (*box.min, 0)

		super().__init__(scene, pts, color)

	
class SplineDisplay(Display):
	''' display for spline curve, with handles around'''
	color: fvec4
	box: Box
	selected: bool
	hovered: bool
	
	def __init__(self, scene, handles, curve, color=None):
		self.color = fvec3(color or settings.colors['line'])
		self.color_handles = fvec3(settings.colors['annotation'])
		self.box = npboundingbox(handles)
		self.selected = False
		self.hovered = False
		
		ctx = scene.context
		vb_handles = ctx.buffer(handles)
		vb_curve = ctx.buffer(curve)
		
		shader = scene.share(load_shader_uniformcolor, load_shader_uniformcolor)
		self._va_handles = ctx.vertex_array(shader, [(vb_handles, '3f4', 'v_position')])
		self._va_curve = ctx.vertex_array(shader, [(vb_curve, '3f4', 'v_position')], mode=mgl.LINE_STRIP)
		
		shader = scene.share(load_shader_ident, load_shader_ident)
		self._va_ident = ctx.vertex_array(shader, [(vb_curve, '3f4', 'v_position')], mode=mgl.LINE_STRIP)
	
	def stack(self, scene):
		return (	(self, 'screen', 1, self._render),
					(self, 'ident', 1, self._identify)	)
	
	def _render(self, view):
		self._va_handles.program['color'].write(fvec4(highlight_color(self, settings.colors['annotation']), 0.6))
		self._va_curve.program['color'].write(fvec4(highlight_color(self, self.color), 1))
		
		self._va_handles.program['view'].write(view.uniforms['view'] * self.world)
		self._va_handles.program['proj'].write(view.uniforms['proj'])
		view.scene.context.point_size = 4
		
		self._va_handles.program['layer'] = -2e-6
		self._va_handles.program['color'].write(fvec4(highlight_color(self, self.color_handles), 0.5))
		self._va_handles.render(mgl.POINTS)
		self._va_handles.render(mgl.LINE_STRIP)
		
		self._va_curve.program['layer'] = -1e-6
		self._va_curve.program['color'].write(fvec4(highlight_color(self, self.color), 1))
		self._va_curve.render()
		
	def _identify(self, view):
		self._va_ident.program['ident'] = view.identstep(1)
		self._va_ident.program['view'].write(view.uniforms['view'] * self.world)
		self._va_ident.program['proj'].write(view.uniforms['proj'])
		self._va_ident.render()
		

def digitfit(n):
	''' return the number of zeros in the digital representation of n '''
	s = 0
	while n:
		if not n%10:	s += 1
		n //=10
	return s

class GridDisplay(Display):
	''' display a grid clipped to the view, helping to appreciate distances
		The grid display distances in the plane of the `center` point
	'''
	center: fvec3
	''' 3d position of the center of the grid '''
	unit: float
	''' factor on the distance between dots '''
	color: fvec3
	''' base color of the grid, with an alpha channel '''
	contrast: float
	''' contrast between the smallest division and the biggest '''
			
	def __init__(self, scene, center=None, unit=1., color=None, contrast=1.8):
		self.unit = unit
		self.center = fvec3(center or 0)
		self.color = fvec4(color) if color else fvec4(settings.colors['point'],1)
		self.contrast = contrast
		
		self._va = scene.share(type(self), self._share)

	def _share(self, scene):
		n = 101
		h = n//2
		pts = np.empty((n,n), dtype='f4, f4, f2')
		for i in range(n):
			for j in range(n):
				pts[i,j] = (i-h, j-h, (1+min(digitfit(i), digitfit(j))))
		
		shader = scene.context.program(
			vertex_shader=open(resourcedir+'/shaders/viewgrid.vert').read(),
			fragment_shader=open(resourcedir+'/shaders/viewgrid.frag').read(),
			)
		vb = scene.context.buffer(pts)
		return scene.context.vertex_array(shader, [(vb, '2f4 f2', 'v_position', 'v_opacity')], mode=mgl.POINTS)
	
	def stack(self, scene):
		return (self, 'screen', 3, self._render),
	
	def _render(self, view):
		center = fvec3( view.uniforms['view'] * self.world * fvec4(self.center,1) )
		if not center.z < 0:	return
		zlog = log(-center.z)/log(10)
		sizelog = floor(zlog)
		
		view.scene.context.point_size = 1/400 * min(view.uniforms['size'])
		self._va.program['color'].write(fvec4(
				self.color.rgb, 
				self.color.a * exp(self.contrast*(-1+sizelog-zlog)),
				))
		self._va.program['size'] = self.unit * 10**(sizelog-1)
		self._va.program['centerdist'] = -center.z
		self._va.program['proj'].write(view.uniforms['proj'])
		self._va.program['contrast'] = self.contrast
		self._va.render()


# class CutPlane(Displau):
#     ''' show a plane cutting all meshes '''


def tupledisplay(scene: Scene, t:tuple) -> Display:
	if primitives.isaxis(t):	return AxisDisplay(scene, t)
	# if not found: empty display
	return Display()

Scene.overrides.update({
	dvec3:   PointDisplay,
	fvec3:  PointDisplay,
	tuple:  tupledisplay,
	Box:    BoxDisplay,
	})



# This file is part of pymadcad,  distributed under license LGPL v3
'''
	implementation of views and displays for 3D
'''
import numpy as np

from ...common import resourcedir
from ...mathutils import Box, fvec3
from ..base import Display

def load_shader_ident(scene):
	return scene.context.program(
		vertex_shader=open(resourcedir+'/shaders/object-ident.vert').read(),
		fragment_shader=open(resourcedir+'/shaders/ident.frag').read(),
		)

def load_shader_subident(scene):
	return scene.context.program(
		vertex_shader=open(resourcedir+'/shaders/object-item-ident.vert').read(),
		fragment_shader=open(resourcedir+'/shaders/ident.frag').read(),
		)        

def load_shader_wire(scene):
	return scene.context.program(
		vertex_shader=open(resourcedir+'/shaders/wire.vert').read(),
		fragment_shader=open(resourcedir+'/shaders/wire.frag').read(),
		)
	
def load_shader_uniformcolor(scene):
	return scene.context.program(
		vertex_shader=open(resourcedir+'/shaders/uniformcolor.vert').read(),
		fragment_shader=open(resourcedir+'/shaders/uniformcolor.frag').read(),
		)

def npboundingbox(points: np.ndarray, ignore=False) -> Box:
	''' boundingbox for numpy arrays of points on the 3 first components '''
	if not ignore:
		return Box(
			fvec3(np.min(points, axis=0)),
			fvec3(np.max(points, axis=0)),
			)
	else:
		return Box(
			fvec3(np.nan_to_num(np.min(points, axis=0), True, +np.inf)),
			fvec3(np.nan_to_num(np.max(points, axis=0), True, -np.inf)),
			)

from . import view, marker, dense
from .view import *



# This file is part of pymadcad,  distributed under license LGPL v3
'''	
	Display module of pymadcad
	
	This module provides a render pipeline system featuring:

	- Class `Scene` to gather the data to render, basically instances of `Display`
	- many kind of `View` that actually renders the scene 
	- The `Display` base class meant to be subclassed into any kind of display
	- The `Group` class that allows to nest displays in the scene

	The `Scene` is only to manage the objects to render . Almost all madcad data types can be rendered to scenes being converted into an appropriate subclass of `Display`. Since the conversion from madcad data types into display instance is automatically handled via the object's `display` method (or scene overrides), you usually don't need to deal with displays directly.
	
	The views can be widgets or simply a bunch of buffers to render the scene onto. There is different view implementations for different use cases.
		
	WARNING
	-------
		As the GPU native precision is f4 (float 32 bits), all the vector stuff regarding rendering is made using simple precision types: `fvec3, fvec4, fmat3, fmat4, ...`
'''
from __future__ import annotations

from .. import settings
from .utils import *
from .base import *
from .d3.view import *
# from .d2.view import *

try:
	from ..qt import QApplication
except ImportError:
	pass
else:

	def show(scene:Scene|dict|list, interest: Box = None, size=uvec2(400, 400), projection=None, navigation=True, **options):
		'''
		Easy and convenient way to create a window containing a `View3D` on a `Scene`

		If a `QApplication` is not already running, a `QApplication` is created and the functions only returns when the window has been closed and all GUI destroyed

		Parameters:
			scene:     a mapping (dict or list) giving the objects to render in the scene
			interest:  the region of interest to zoom on at the window initialization
			size:      the window size (pixel)
			projection: an object handling the camera projection (aka intrinsic parameters), see `QView3D.projection`
			navigation: an object handling the camera movements, see `QView3D.navigation`
			options:   options to set in `Scene.options`

		Tip:
			For integration in a Qt window or to manipulate the view, you should directly use `View`
		'''
		import sys
		
		if not isinstance(scene, Scene):
			scene = Scene(scene, options)

		# detect existing QApplication
		app = QApplication.instance()
		created = False
		if not app:
			app = QApplication(sys.argv)
			created = True

		# use the Qt color scheme if specified
		if settings.display['system_theme']:
			settings.use_qt_colors()

		# create the scene as a window
		view = QView3D(scene, projection=projection, navigation=navigation)
		view.resize(*size)
		view.show()
		scene.prepare()

		# make the camera see everything
		if not interest:	
			interest = view.scene.root.box
		view.center(interest.center)
		view.adjust(interest)

		# run eventually created QApplication
		if created:
			err = app.exec()
			if err != 0:
				print('error: Qt exited with code', err)


def render(scene:Scene|dict|list, size=uvec2(400, 400), view:fmat4=None, proj:fmat4=None, interest:Box=None, alpha=False, **options) -> ndarray:
	'''
	render the given scene to an image

	The output image is a numpy tensor RGB8 (w,h,3) or RGBA8 (w,h,4)
	For repeated renderings or view manipulation, you should directly use `Offscreen3D` rather than calling this function repeatedly

	NOTE:
		The system theme colors cannot be automatically loaded since no running QApplication is assumed in the function
		
	Parameters:
		scene:     a mapping (dict or list) giving the objects to render in the scene
		interest:  the region of interest to zoom on at the window initialization
		size:      the resulting image size (pixel)
		view:    the view matrix, which should be a affine matrix with rotation and translation
		proj:    the openGL projection matrix, could be generated with `perspective` or `orthographic`		
		alpha:   whether to add an alpha channel for the background in the resulting image
	'''
	scene = Scene(scene, options)
	if not interest:
		interest = scene.root.box
	size = length(interest.width)
	if not proj:
		proj = globals()[settings.scene['projection']]().matrix(size)
	if not view:
		view = translate(interest.center + size*Z)
		# TODO take the projection transformation into account to get an appropriate distance
	view = Offscreen3D(scene, size, view, proj, enable_alpha=alpha)
	scene.prepare()
	return view.render()



# This file is part of pymadcad,  distributed under license LGPL v3
import numpy as np
import moderngl as mgl
from PIL import Image

from .. import settings
from ..common import resourcedir
from ..mathutils import *
from ..rendering import Display
from ..rendering.d3 import npboundingbox, load_shader_wire, load_shader_subident

__all__ = ['MeshDisplay', 'WebDisplay']


class MeshDisplay(Display):
	''' Display render Meshes '''
	def __init__(self, scene, positions, normals, faces, lines, idents, color=None):
		self.box = npboundingbox(positions, ignore=True)
		
		color = fvec3(color or settings.colors['surface'])
		line = ( (length(settings.colors['line']) + dot(
					color - settings.colors['surface'], 
					settings.colors['line'] - settings.colors['surface']))
				* normalize(color + 1e-6)  )
		#if length(s['line_color']) > length(color)
		reflect = normalize(color + 1e-6) * settings.display['solid_reflectivity']
		
		self._vertices = Vertices(scene.context, positions, idents)
		self._disp_faces = FacesDisplay(scene, self._vertices, normals, faces, color=color, reflect=reflect, layer=0)
		self._disp_ghost = GhostDisplay(scene, self._vertices, normals, faces, color=line, layer=0)
		self._disp_groups = LinesDisplay(scene, self._vertices, lines, color=line, alpha=1, layer=-2e-6)
		self._disp_points = PointsDisplay(scene, self._vertices, range(len(positions)), color=line, layer=-3e-6)
		wire = []
		for f in faces:
			wire.append((f[0], f[1]))
			wire.append((f[1], f[2]))
			wire.append((f[2], f[0]))
		self._disp_wire = LinesDisplay(scene, self._vertices, wire, color=line, alpha=0.3, layer=-1e-6)
		
	def stack(self, scene):
		yield (self, 'screen', -1, self._vertices.prerender)
		if scene.options['display_faces']:	
			yield (self, 'screen', 0, self._disp_faces.render)
			yield (self, 'ident', 0, self._disp_faces.identify)
		else:								
			yield (self, 'screen', 1, self._disp_ghost.render)
			yield (self, 'ident', 0, self._disp_ghost.identify)
		if scene.options['display_groups']:	yield (self, 'screen', 1, self._disp_groups.render)
		if scene.options['display_points']:	yield (self, 'screen', 2, self._disp_points.render)
		if scene.options['display_wire']:	yield (self, 'screen', 2, self._disp_wire.render)
	
	def _get_world(self):	
		return self._vertices.world
	def _set_world(self, value):
		self._vertices.world = value
	world:fmat4 = property(_get_world, _set_world)
	
	def _get_selected(self):
		return self._vertices.selected
	def _set_selected(self, value):
		if not isinstance(value, set):
			if value:
				self._vertices.selected.add(None)
			else:
				self._vertices.selected.clear()
			value = self._vertices.selected
		self._vertices.selected = value
		self._vertices.flags_updated = True
	selected:set = property(_get_selected, _set_selected)
	
	def _get_hovered(self):
		return self._vertices.hovered
	def _set_hovered(self, value):
		if not isinstance(value, set):
			if value:
				self._vertices.hovered.add(None)
			else:
				self._vertices.hovered.clear()
			value = self._vertices.hovered
		self._vertices.hovered = value
		self._vertices.flags_updated = True
	hovered:set = property(_get_hovered, _set_hovered)
	

class WebDisplay(Display):
	''' Display to render Webs '''
	def __init__(self, scene, positions, lines, points, idents, color=None):
		self.box = npboundingbox(positions, ignore=True)
		color = color or settings.colors['line']
		self._vertices = Vertices(scene.context, positions, idents)
		self._disp_edges = LinesDisplay(scene, self._vertices, lines, color=color, alpha=1, layer=-2e-6)
		self._disp_groups = PointsDisplay(scene, self._vertices, points, layer=-3e-6)
		self._disp_points = PointsDisplay(scene, self._vertices, range(len(positions)), layer=-1e-6)

	def stack(self, scene):
		yield (self, 'screen', -1, self._vertices.prerender)
		if scene.options['display_groups']:		yield (self, 'screen', 2, self._disp_groups.render)
		if scene.options['display_points']:		yield (self, 'screen', 2, self._disp_points.render)
		yield (self, 'screen', 1, self._disp_edges.render)
		yield (self, 'ident', 1, self._disp_edges.identify)
	
	def _get_world(self):	
		return self._vertices.world
	def _set_world(self, value):	
		self._vertices.world = value
	world = property(_get_world, _set_world)
	
	def _get_selected(self):
		return self._vertices.selected
	def _set_selected(self, value):
		if not isinstance(value, set):
			if value:
				self._vertices.selected.add(None)
			else:
				self._vertices.selected.clear()
			value = self._vertices.selected
		self._vertices.selected = value
		self._vertices.flags_updated = True
	selected:set = property(_get_selected, _set_selected)
	
	def _get_hovered(self):
		return self._vertices.hovered
	def _set_hovered(self, value):
		if not isinstance(value, set):
			if value:
				self._vertices.hovered.add(None)
			else:
				self._vertices.hovered.clear()
			value = self._vertices.hovered
		self._vertices.hovered = value
		self._vertices.flags_updated = True
	hovered:set = property(_get_hovered, _set_hovered)


class Vertices:
	''' convenient class to share vertices between SolidDisplay, WebDisplay, PointsDisplay '''
	# vertex flags for shaders
	HOVERED = 1<<0
	SELECTED = 1<<1
	
	def __init__(self, ctx, positions, idents):
		self.idents = idents
		self.nident = int(max(idents))+1
		self.u_flags = 0
		self.v_flags = np.zeros(len(positions), dtype='u1')
		self.flags_updated = False
		self.selected = set()
		self.hovered = set()
		
		assert len(idents) == len(positions)
		self.vb_positions = ctx.buffer(np.asarray(positions, dtype='f4', order='C'))
		self.vb_idents = ctx.buffer(np.asarray(idents, dtype='u2', order='C'))
		self.vb_flags = ctx.buffer(self.v_flags, dynamic=True)
		self.world = fmat4(1)
		
	def prerender(self, view):
		if self.flags_updated:
			self.flags_updated = False
			self.u_flags = (
				(None in self.hovered)*self.HOVERED
				| (None in self.selected)*self.SELECTED
				)
			if (None not in self.selected) or (None not in self.hovered):
				for i,ident in enumerate(self.idents):
					self.v_flags[i] = (
						(ident in self.hovered)*self.HOVERED
						| (ident in self.selected)*self.SELECTED
						)
				self.vb_flags.write(self.v_flags)


class FacesDisplay:
	def __init__(self, scene, vertices, normals, faces, color, reflect, layer=0):
		self.color = color
		self.layer = layer
		self.reflect = reflect
		self._vertices = vertices
	
		# load the skybox texture
		def load(scene):
			img = Image.open(resourcedir+'/textures/'+settings.display['solid_reflect'])
			return scene.context.texture(img.size, 3, img.tobytes())
		self.reflectmap = scene.share('skybox', load)
		
		# load the shader
		self.shader = scene.share(type(self), self._share)
		self.ident_shader = scene.share(load_shader_subident, load_shader_subident)
		# allocate buffers
		if faces is not None and len(faces) and vertices.vb_positions:
			self.vb_faces = scene.context.buffer(np.asarray(faces, 'u4', order='C'))
			self.vb_normals = scene.context.buffer(np.asarray(normals, 'f4', order='C'))
			self.va = scene.context.vertex_array(
					self.shader, 
					[	(vertices.vb_positions, '3f', 'v_position'), 
						(self.vb_normals, '3f', 'v_normal'),
						(vertices.vb_flags, 'u1', 'v_flags')],
					self.vb_faces,
					mode=mgl.TRIANGLES,
					)
			
			self.va_ident = scene.context.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_faces,
					mode=mgl.TRIANGLES,
					)
		else:
			self.va = None
	
	def _share(self, scene):
		shader = scene.context.program(
			vertex_shader=open(resourcedir+'/shaders/solid.vert').read(),
			fragment_shader=open(resourcedir+'/shaders/solid.frag').read(),
			)
		# setup some uniforms
		shader['reflectmap'] = 0
		return shader
	
	def render(self, view):
		if self.va:
			self.reflectmap.use(0)
			self.va.program['u_flags'] = self._vertices.u_flags
			self.va.program['selected_color'].write(settings.display['selection_color'])
			self.va.program['hovered_color'].write(settings.display['hover_color'])
			self.va.program['min_color'].write(self.color * settings.display['solid_color_side'])
			self.va.program['max_color'].write(self.color * settings.display['solid_color_front'])
			self.va.program['refl_color'].write(self.reflect)
			self.va.program['layer'] = self.layer
			self.va.program['world'].write(self._vertices.world)
			self.va.program['view'].write(view.uniforms['view'])
			self.va.program['proj'].write(view.uniforms['proj'])
			self.va.render()
	
	def identify(self, view):
		if self.va:
			self.va_ident.program['layer'] = self.layer
			self.va_ident.program['start_ident'] = view.identstep(self._vertices.nident)
			self.va_ident.program['view'].write(view.uniforms['view'] * self._vertices.world)
			self.va_ident.program['proj'].write(view.uniforms['proj'])
			self.va_ident.render()

class GhostDisplay:
	def __init__(self, scene, vertices, normals, faces, color, layer=0):
		self.color = color
		self.layer = layer
		self._vertices = vertices
		
		self.shader = scene.share(type(self), self._share)
		self.ident_shader = scene.share(load_shader_subident, load_shader_subident)
		# allocate buffers
		if faces is not None and len(faces) and vertices.vb_positions:
			self.vb_faces = scene.context.buffer(np.asarray(faces, 'u4', order='C'))
			self.vb_normals = scene.context.buffer(np.asarray(normals, 'f4', order='C'))
			self.va = scene.context.vertex_array(
					self.shader, 
					[	(vertices.vb_positions, '3f', 'v_position'), 
						(self.vb_normals, '3f', 'v_normal'),
						(vertices.vb_flags, 'u1', 'v_flags')],
					self.vb_faces,
					mode=mgl.TRIANGLES,
					)
			
			self.va_ident = scene.context.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_faces,
					mode=mgl.TRIANGLES,
					)
		else:
			self.va = None
	
	def _share(self, scene):
		return scene.context.program(
			vertex_shader=open(resourcedir+'/shaders/solid.vert').read(),
			fragment_shader=open(resourcedir+'/shaders/ghost.frag').read(),
			)
	
	def render(self, view):
		if self.va:
			# setup uniforms
			view.scene.context.enable_only(mgl.BLEND)
			self.va.program['u_flags'] = self._vertices.u_flags
			self.va.program['normal_color'].write(self.color)
			self.va.program['selected_color'].write(settings.display['selection_color'])
			self.va.program['hovered_color'].write(settings.display['hover_color'])
			self.va.program['world'].write(self._vertices.world)
			self.va.program['view'].write(view.uniforms['view'])
			self.va.program['proj'].write(view.uniforms['proj'])
			self.va.program['layer'] = self.layer
			self.va.render()
			view.scene.context.enable_only(mgl.BLEND | mgl.DEPTH_TEST)
	
	def identify(self, view):
		if self.va:
			self.ident_shader['start_ident'] = view.identstep(self._vertices.nident)
			self.ident_shader['view'].write(view.uniforms['view'] * self._vertices.world)
			self.ident_shader['proj'].write(view.uniforms['proj'])
			self.ident_shader['layer'] = self.layer
			# render on self.context
			self.va_ident.render()

class LinesDisplay:
	def __init__(self, scene, vertices, lines, color, alpha=1, layer=0):
		self.layer = layer
		self.color = fvec4(fvec3(color), alpha)
		self._vertices = vertices
		
		# load the line shader
		self.shader = scene.share(load_shader_wire, load_shader_wire)
		self.ident_shader = scene.share(load_shader_subident, load_shader_subident)
		if lines is not None and len(lines) and vertices.vb_positions:
			# allocate buffers
			self.vb_lines = scene.context.buffer(np.asarray(lines, dtype='u4', order='C'))
			self.va = scene.context.vertex_array(
						self.shader,
						[	(vertices.vb_positions, '3f', 'v_position'),
							(vertices.vb_flags, 'u1', 'v_flags')],
						self.vb_lines,
						mode = mgl.LINES,
						)
			self.va_ident = scene.context.vertex_array(
					self.ident_shader, 
					[	(vertices.vb_positions, '3f', 'v_position'),
						(vertices.vb_idents, 'u2', 'item_ident')], 
					self.vb_lines,
					mode = mgl.LINES,
					)
		else:
			self.va = None
	
	def render(self, view):
		if self.va:
			self.va.program['u_flags'] = self._vertices.u_flags
			self.va.program['color'].write(self.color)
			self.va.program['selected_color'].write(settings.display['selection_color'])
			self.va.program['hovered_color'].write(settings.display['hover_color'])
			self.va.program['view'].write(view.uniforms['view'] * self._vertices.world)
			self.va.program['proj'].write(view.uniforms['proj'])
			self.va.program['layer'] = self.layer
			self.va.render()
		
	def identify(self, view):
		if self.va:
			self.va_ident.program['start_ident'] = view.identstep(self._vertices.nident)
			self.va_ident.program['view'].write(view.uniforms['view'] * self._vertices.world)
			self.va_ident.program['proj'].write(view.uniforms['proj'])
			self.va_ident.program['layer'] = self.layer
			self.va_ident.render()
		
class PointsDisplay:
	def __init__(self, scene, vertices, indices=None, color=None, ptsize=3, layer=0):
		self.color = fvec4(color or settings.colors['point'], 1)
		self.select_color = fvec4(settings.display['selection_color'].rgb, 1)
		self.ptsize = ptsize
		self.layer = layer
		self._vertices = vertices
		
		# load the line shader
		self.shader = scene.share(load_shader_wire, load_shader_wire)
		self.ident_shader = scene.share(load_shader_subident, load_shader_subident)
		# allocate GPU objects
		if indices is not None and len(indices) and vertices.vb_positions:
			self.vb_indices = scene.context.buffer(np.asarray(indices, dtype='u4', order='C'))
			self.va = scene.context.vertex_array(
				self.shader,
				[	(vertices.vb_positions, '3f', 'v_position'),
					(vertices.vb_flags, 'u1', 'v_flags')],
				self.vb_indices,
				mode=mgl.POINTS,
				)
			self.va_ident = scene.context.vertex_array(
				self.ident_shader, 
				[	(vertices.vb_positions, '3f', 'v_position'),
					(vertices.vb_idents, 'u2', 'item_ident')], 
				self.vb_indices,
				mode=mgl.POINTS,
				)
		else:
			self.va = None
	
	def render(self, view):
		if self.va:
			view.scene.context.point_size = self.ptsize
			self.va.program['u_flags'] = self._vertices.u_flags
			self.va.program['layer'] = self.layer
			self.va.program['color'].write(self.color)
			self.va.program['selected_color'].write(settings.display['selection_color'])
			self.va.program['hovered_color'].write(settings.display['hover_color'])
			self.va.program['view'].write(view.uniforms['view'] * self._vertices.world)
			self.va.program['proj'].write(view.uniforms['proj'])
			self.va.render()
	
	def identify(self, view):
		if self.va:
			view.ctx.point_size = self.ptsize
			self.va_ident.program['layer'] = self.layer
			self.va_ident.program['start_ident'] = view.identstep(self._vertices.nident)
			self.va_ident.program['view'].write(view.uniforms['view'] * self._vertices.world)
			self.va_ident.program['proj'].write(view.uniforms['proj'])
			self.va_ident.render()



# This file is part of pymadcad,  distributed under license LGPL v3
'''
	module providing different kind of 3d views over a Scene, and few helpers with them
'''
from __future__ import annotations

import numpy as np
import moderngl as mgl
from pnprint import nprint

from ... import settings
from ...mathutils import *
from ..base import Scene, empty
from ..utils import *

__all__ = ['GLView3D', 'Offscreen3D', 'QView3D', 'Orbit', 'Turntable', 'Perspective', 'Orthographic']


class GLView3D:
	''' pure openGL 3D view over a madcad scene '''
	scene: Scene
	''' madcad scene to render, it must contain 3D displays (at least a the root) '''
	screen: mgl.Framebuffer
	''' color map (and depth) openGL buffer '''
	ident: mgl.Framebuffer|None
	''' identification map (and depth) openGL buffer '''
	targets: list
	''' exposed for `Scene` '''
	uniforms: dict
	''' exposed for `Scene` '''
	view: fmat4
	''' current view matrix, this will be the default for next rendering '''
	proj: fmat4
	''' current projection matrix, this will be the default for next rendering '''
	enable_ident: bool
	''' enable `self.ident` '''
	
	def __init__(self, scene, size:uvec2=None, view:fmat4=None, proj:fmat4=None, 
			enable_ident=True, 
			**uniforms):
		self.scene = scene if isinstance(scene, Scene) else Scene(scene)
		self.view = view
		self.proj = proj
		self.uniforms = uniforms
		self.screen = None
		self.ident = None
		self.targets = []
		self.enable_ident = enable_ident
		# indentifiers management for the ident map
		self._steps = []  # identifiers before each ident stack step
		self._stepi = 0
		self._step = 0
		
		self.scene.root.world = fmat4()
		if size:
			self._reallocate(size)
	
	def render(self, size:uvec2=None, view:fmat4=None, proj:fmat4=None) -> Self:
		''' trigger the rendering of a frame, do not wait for the result 
		
			- the `view` and `proj` instance attributes can be changed on the fly without extra cost.
			- a `size` change will trigger reallocation of the buffers
		'''
		size = glsize(size)
		self._reallocate(size)
		# setup uniforms
		if view:    self.view = view
		if proj:    self.proj = proj
		self.uniforms['size'] = uvec2(self.screen.size)
		self.uniforms['view'] = self.view
		self.uniforms['proj'] = self.proj
		self.uniforms['projview'] = self.proj * self.view
		# normal rendering of the scene
		self.scene.render(self)
		# downsample the screen buffer to 1 sample per pixel
		self.scene.context.copy_framebuffer(self.screen, self._screen_multisample)
		
		# self.screen.use()
		# self.screen.clear(1,0,0)
		self.scene.context.finish()
		return self
		
	def _reallocate(self, size):
		''' ensure the framebuffer fit the given size, reallocate if needed '''
		if self.screen and size == self.screen.size:
			return
		ctx = self.scene.context
		self.targets.clear()
		self.screen = ctx.simple_framebuffer(size, components=3, dtype='f1')
		self._screen_multisample = ctx.simple_framebuffer(size, components=3, samples=4, dtype='f1')
		self.targets.append(('screen', self._screen_multisample, self._setup_screen))
		if self.enable_ident:
			self.ident = ctx.simple_framebuffer(size, components=1, dtype='u2')
			self.targets.append(('ident', self.ident, self._setup_ident))
		else:
			self.ident = None

	def _setup_ident(self):
		''' ident rendering setup '''
		self._stepi = 0
		self._step = 1
		if len(self.scene.stacks.get('ident', empty)) != len(self._steps):
			self._steps = [0] * len(self.scene.stacks['ident'])
		# ident rendering setup
		ctx = self.scene.context
		ctx.multisample = False
		ctx.enable_only(mgl.DEPTH_TEST)
		ctx.blend_func = mgl.ONE, mgl.ZERO
		ctx.blend_equation = mgl.FUNC_ADD
		self.target.clear()

	def _setup_screen(self):
		''' screen rendering setup '''
		ctx = self.scene.context
		ctx.multisample = True
		ctx.enable_only(mgl.BLEND | mgl.DEPTH_TEST)
		ctx.blend_func = mgl.SRC_ALPHA, mgl.ONE_MINUS_SRC_ALPHA
		ctx.blend_equation = mgl.FUNC_ADD
		
		background = settings.display['background_color']
		if len(background) == 3:
			self.target.clear(*background, alpha=1)
		elif len(background) == 4:
			self.target.clear(*background)
		else:
			raise ValueError(f"background_color must be a RGB or RGBA tuple, currently {background}")

	def identstep(self, nidents):
		''' Updates the amount of rendered idents and return the start ident for the calling rendering pass?
			Method to call during a renderstep
		'''
		s = self._step
		self._step += nidents
		self._steps[self._stepi] = self._step-1
		self._stepi += 1
		return s

	
class Offscreen3D:
	''' 3D view giving images accessible to numpy buffers '''
	gl: View
	''' opengl renderer '''
	color: ndarray[np.uint8]
	''' result image from the previous rendering, showing the  '''
	depth: ndarray[np.float32]|None
	''' result image from the previous rendering '''
	ident: ndarray[np.uint16]|None
	''' result image from the previous rendering '''
	enable_alpha: bool
	''' if enabled, `self.color` is RGBA else it is 'RGB' '''
	enable_depth: bool
	''' enable `self.depth` '''
	enable_ident: bool
	''' enable `self.ident` '''
		
	scene = forwardproperty('gl', 'scene')
	view = forwardproperty('gl', 'view')
	proj = forwardproperty('gl', 'proj')
	uniforms = forwardproperty('gl', 'uniforms')
	
	def __init__(self, scene, size:uvec2=None, view:fmat4=None, proj:fmat4=None, 
			enable_depth=False, 
			enable_ident=False, 
			enable_alpha=False,
			**uniforms):
		self.gl = GLView3D(size, view, proj, enable_ident, **uniforms)
		self.color = self.depth = self.ident = None
		self.enable_depth = enable_depth
		self.enable_alpha = enable_alpha
		if size:
			self._reallocate(size)
	
	def render(self, size:uvec2=None, view:fmat4=None, proj:fmat4=None) -> Self:
		'''
			render the scene and retreive the result in the `color`, `depth` and `ident` attributes
			
			- the `view` and `proj` instance attributes can be changed on the fly without extra cost.
			- a `size` change will trigger reallocation of the buffers
		'''
		size = glsize(size)
		self.gl.render(size, view, proj)
		self._reallocate(size)
		# retreive everything from the GPU to the CPU
		if self.enable_alpha:
			self.gl.screen.read_into(self.color, attachments=4)
		else:
			self.gl.screen.read_into(self.color, attachments=3)
		# switch vertical axis to convert from opengl image convention to usual (and Qt) image convention
		self.color[:] = self.color[::-1]
		if self.enable_ident:
			self.gl.ident.read_into(self.ident, attachment=0)
			self.ident[:] = self.ident[::-1]
		if self.enable_depth:
			self.gl.screen.read_into(self.depth, attachment=-1)
			self.depth[:] = self.depth[::-1]
		return self
		
	def _reallocate(self, size):
		if self.colr is not None and size == self.color.shape[::-1]:
			return
		w,h = size
		self.color = np.empty((h,w,3), dtype='u1')
		self.ident = np.empty((h,w), dtype='u2')  if self.enable_ident else None
		self.depth = np.empty((h,w), dtype='f4')  if self.enable_depth else None


try:
	from ...qt import (Qt, QApplication, QWidget, QImage, QPainter, 
					QEvent, QInputEvent, QMouseEvent, QKeyEvent, QTouchEvent)
except ImportError:
	pass
else:

	class QView3D(QWidget):
		gl: View
		''' underlying opengl rendering system '''
		navigation: Orbit|TurnTable|None
		''' turns basic screen actions into camera movements '''
		projection: Perspective|Orthographic
		''' generates the projection matrix '''
		color: ndarray[np.uint8]
		''' last rendering '''
		depth: CheapMap[np.float32]
		''' access to the depthmap from last rendering '''
		ident: CheapMap[np.uint16]
		''' access to the identification map from last rendering '''
		callbacks: list
		''' list of callbacks subscribing to input events
		
			they will be called in revered order, and as any Qt event handler, this first accepting prevents the other to be called
		'''
		
		scene = forwardproperty('gl', 'scene')
		view = forwardproperty('gl', 'view')
		proj = forwardproperty('gl', 'proj')
		uniforms = forwardproperty('gl', 'uniforms')
		
		def __init__(self, scene, projection=None, navigation=True, parent=None):
			QWidget.__init__(self, parent)
			self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
			self.setAttribute(Qt.WidgetAttribute.WA_AcceptTouchEvents, True)
			self.setMouseTracking(True)
			self.gl = GLView3D(scene)
			if navigation is True:
				navigation = globals()[settings.controls['navigation']]()
			self.navigation = navigation
			self.projection = projection or globals()[settings.scene['projection']]()
			self.color = self.depth = self.ident = None
			self.callbacks = []
			
			self._move_mode = receiver(self._move_mode())
			self._keyboard_move = receiver(self._keyboard_move())
			self._mouse_move = receiver(self._mouse_move())
			self._touch_move = receiver(self._touch_move())
			
		def _reallocate(self, size):
			w, h = size
			if self.color is not None and w == self.color.shape[1] and h == self.color.shape[0]:
				return
			self.ident = CheapMap(self.gl.ident, attachment=0)
			self.depth = CheapMap(self.gl.ident, attachment=-1)
			self.color = np.empty((h,w,3), dtype='u1')
		
		def paintEvent(self, painter):
			w = self.width()
			h = self.height()
			size = glsize(uvec2(w, h))
			
			self.gl.render(
				size,
				view = self.navigation.matrix(),
				proj = self.projection.matrix(w/h if h > 0 else 0, self.navigation.distance),
			)
			self._reallocate(size)
			# retreive buffer
			self.gl.screen.read_into(self.color, components=3)
			# switch vertical axis to convert from opengl image convention to usual (and Qt) image convention
			self.color[:] = self.color[::-1]
			# copy to Qt
			image = QImage(
				self.color,
				self.color.shape[1], 
				self.color.shape[0],
				QImage.Format.Format_RGB888)
			painter = QPainter(self).drawImage(0, 0, image)
		
		def event(self, evt):
			if isinstance(evt, QInputEvent):
				evt.ignore()
				self.inputEvent(evt)
				if evt.isAccepted():	
					return True
			return super().event(evt)
		
		def resizeEvent(self, evt):
			QWidget.resizeEvent(self, evt)
			self.update()
		
		def changeEvent(self, evt):
			# detect theme change
			if evt.type() == QEvent.Type.PaletteChange and settings.display['system_theme']:
				settings.use_qt_colors()
			return QWidget.changeEvent(self, evt)
		
		def leaveEvent(self, evt):
			self.scene.hover = None
			self.update()
		
		def inputEvent(self, evt):
			''' Default handler for every input event (mouse move, press, release, keyboard, ...)
				When the event is not accepted, the usual matching Qt handlers are used (mousePressEvent, KeyPressEvent, etc).

				This function can be overwritten to change the view widget behavior.
			'''
			evt.ignore()
			
			for i in reversed(range(len(self.callbacks))):
				if not self.callbacks[i](evt):
					self.callbacks.pop(i)
				if evt.isAccepted():
					return
			
			if self.navigation:
				self._move_mode(evt)
				self._keyboard_move(evt)
				self._mouse_move(evt)
				self._touch_move(evt)
				if evt.isAccepted():
					return
			
			# send the event to the scene objects, descending the item tree
			pos = None
			if isinstance(evt, QMouseEvent):
				pos = self.somenear(evt.pos())
				if pos:
					key = self.displayat(pos)
					if key:
						disp, sub = key
						key = (*disp.key, sub)
						
						self.control(key, evt)
						if evt.isAccepted():
							return
						self._selection(disp, sub, evt)
						if evt.isAccepted():
							return
			if not pos:
				self._deselection(evt)
			
			# if clicks are not accepted, then some following keyboard events may not come to the widget
			# NOTE this also discarding the ability to move the window from empty areas
			if evt.type() == QEvent.Type.MouseButtonPress:
				evt.accept()
		
		def control(self, key, evt):
			''' Transmit a control event successively to all the displays matching the key path stages.
				At each level, if the event is not accepted, it transmits to sub items

				This function can be overwritten to change the interaction with the scene objects.
			'''
			disp = self.scene.root.displays
			for i in range(1,len(key)):
				disp = disp[key[i-1]]
				disp.control(self, key[:i], key[i:], evt)
				if evt.isAccepted(): 
					self.update()
					break
				
		def _selection(self, disp, sub, evt):
			''' handle the selection events '''
			if not self.scene.options['selection_sub']:
				sub = None
			
			# selection on click above
			if evt.type() == QEvent.Type.MouseButtonRelease and evt.button() == Qt.MouseButton.LeftButton and hasattr(disp, 'selected'):
				shift = bool(QApplication.queryKeyboardModifiers() & Qt.KeyboardModifier.ShiftModifier)
				if not self.scene.options['selection_multiple'] ^ shift:
					self.scene.selection_clear()
				self.scene.selection_toggle(disp, sub)
				self.update()
				evt.accept()
				return
			
			# hover on move above
			if evt.type() == QEvent.Type.MouseMove and hasattr(disp, 'hovered'):
				# precise condition on hover change, to avoid triggering unnecessary renderings
				if disp is not self.scene.hover or (isinstance(disp.hovered, set) and sub not in disp.hovered):
					self.scene.hover = (disp, sub)
					self.update()
					
		def _deselection(self, evt):
			''' handle the deselection events '''
			if evt.type() == QEvent.Type.MouseButtonPress and evt.button() == Qt.MouseButton.LeftButton:
				self.scene.selection_clear()
				self.update()
				evt.accept()
			elif evt.type() == QEvent.Type.MouseMove:
				if self.scene.hover:
					self.scene.hover = None
					self.update()
		
		# move the view following events
		# navigation modes
		_ZOOM = 1
		_PAN = 2
		_ROTATE = 3
		
		def _move_mode(self):
			''' change the navigation mode according to keyboard or mouse events '''
			ctrl = alt = False
			self._slow = False
			self._mode = None
			while True:
				evt = yield
				
				if isinstance(evt, QKeyEvent):
					k = evt.key()
					press = evt.type() == QEvent.Type.KeyPress
					if	 k == Qt.Key.Key_Control:	ctrl = press
					elif k == Qt.Key.Key_Alt:		alt = press
					elif k == Qt.Key.Key_Shift:		self._slow = press
				elif evt.type() == QEvent.Type.MouseButtonPress:
					modifiers = QApplication.queryKeyboardModifiers()
					ctrl = bool(modifiers & Qt.KeyboardModifier.ControlModifier)
					alt = bool(modifiers & Qt.KeyboardModifier.AltModifier)
					shift = bool(modifiers & Qt.KeyboardModifier.ShiftModifier)
				else:
					continue
				
				if evt.type() == QEvent.Type.MouseButtonPress and evt.button() == Qt.MouseButton.MiddleButton:
					self._mode = self._ROTATE
				else:
					if ctrl and alt:		self._mode = self._ZOOM
					elif ctrl:				self._mode = self._PAN
					elif alt:				self._mode = self._ROTATE
					else:					self._mode = None
					
		def _keyboard_move(self):
			''' navigate the view with the keyboard '''
			while True:
				evt = yield
				
				if self._mode and evt.type() == QEvent.Type.KeyPress:
					dx = dy = 0
					speed = pi/72 if self._slow else pi/24
					if evt.key() == Qt.Key.Key_Left:     dx = -speed
					elif evt.key() == Qt.Key.Key_Right:  dx = +speed
					elif evt.key() == Qt.Key.Key_Down:   dy = -speed
					elif evt.key() == Qt.Key.Key_Up:     dy = +speed
					else:
						continue
					
					if self._mode == self._PAN:
						self.navigation.pan(fvec2(dx, dy))
					elif self._mode == self._ROTATE:
						self.navigation.rotate(fvec3(dx, dy, 0))
					elif self._mode == self._ZOOM:
						self.navigation.rotate(dy)
						
					evt.accept()
					self.update()
				
		def _mouse_move(self):
			''' navigate the view using the mouse '''
			while True:
				evt = yield
				
				if evt.type() == QEvent.Type.Wheel:
					self.navigation.zoom(exp(-evt.angleDelta().y()/(8*90)))	# the 8 factor is there because of the Qt documentation
					evt.accept()
					self.update()
				
				if self._mode and evt.type() == QEvent.Type.MouseButtonPress:
					last = evt.pos()
					evt.accept()
					while True:
						evt = yield
						
						if not self._mode or evt.type() == QEvent.Type.MouseButtonRelease or isinstance(evt, QMouseEvent) and evt.buttons() == Qt.MouseButton.NoButton:
							break
						if evt.type() != QEvent.Type.MouseMove:
							continue
						
						gap = evt.pos() - last
						dx = gap.x()/self.height()
						dy = gap.y()/self.height()
						if self._mode == self._PAN:		
							self.navigation.pan(fvec2(dx, dy))
						elif self._mode == self._ROTATE:
							self.navigation.rotate(fvec3(dx, dy, 0))
						elif self._mode == self._ZOOM:
							middle = QPoint(self.width(), self.height())/2
							f = (	(last-middle).manhattanLength()
								/	(evt.pos()-middle).manhattanLength()	)
							self.navigation.zoom(f)
						last = evt.pos()
						
						evt.accept()
						self.update()
		
		def _touch_move(self):
			''' navigate the view using the touchscreen '''
			while True:
				evt = yield
				if isinstance(evt, QTouchEvent):
					self._mode = None
					pts = evt.touchPoints()
					# view rotation
					if len(pts) == 2:
						startlength = (pts[0].lastPos()-pts[1].lastPos()).manhattanLength()
						zoom = startlength / (pts[0].pos()-pts[1].pos()).manhattanLength()
						displt = (	(pts[0].pos()+pts[1].pos()) /2 
								-	(pts[0].lastPos()+pts[1].lastPos()) /2 ) /self.height()
						dc = pts[0].pos() - pts[1].pos()
						dl = pts[0].lastPos() - pts[1].lastPos()
						rot = atan2(dc.y(), dc.x()) - atan2(dl.y(), dl.x())
						self.navigation.zoom(zoom)
						self.navigation.rotate(fvec3(displt.x(), displt.y(), rot))
						self.update()
						evt.accept()
					# view translation
					elif len(pts) == 3:
						lc = (	pts[0].lastPos() 
							+	pts[1].lastPos() 
							+	pts[2].lastPos() 
							)/3
						lr = (	(pts[0].lastPos() - lc) .manhattanLength()
							+	(pts[1].lastPos() - lc) .manhattanLength()
							+	(pts[2].lastPos() - lc) .manhattanLength()
							)/3
						cc = (	pts[0].pos() 
							+	pts[1].pos() 
							+	pts[2].pos() 
							)/3
						cr = (	(pts[0].pos() - cc) .manhattanLength()
							+	(pts[1].pos() - cc) .manhattanLength()
							+	(pts[2].pos() - cc) .manhattanLength()
							)/3
						zoom = lr / cr
						displt = (cc - lc)  /self.height()
						self.navigation.zoom(zoom)
						self.navigation.pan(fvec2(displt.x(), displt.y()))
						self.update()
						evt.accept()
					# finish a gesture
					elif evt.type() in (QEvent.Type.TouchEnd, QEvent.Type.TouchUpdate):
						evt.accept()
		
		# forward access to items
				
		def __getitem__(self, key):
			self.root[key]
		def __setitem__(self, key, value):
			self.root[key] = value
		def __iter__(self):
			return iter(self.root)
		
		# extract informations from the ident map
		
		def somenear(self, point: QPoint, radius=None) -> QPoint|None:
			''' Return the closest coordinate to coords, (within the given radius) for which there is an object at
				So if objnear is returning something, objat and ptat will return something at the returned point
			'''
			point = qpoint_to_vec(point)
			if radius is None:
				radius = settings.controls['snap_dist']
			region = self.ident.region((*(point-radius), *(point+radius+1)))
			for x,y in snailaround(uvec2(radius), region.shape, radius):
				ident = int(region[y, x])
				if ident:
					return vec_to_qpoint(point-radius + ivec2(x,y))

		def ptat(self, point: QPoint) -> fvec3|None:
			''' Return the point of the rendered surfaces that match the given window coordinates '''
			point = qpoint_to_vec(point)
			region = self.depth.region((*point, *(point+1)))
			size = self.uniforms['size']
			depthred = float(region[0,0])
			x =  (point.x/size.x *2 -1)
			y = -(point.y/size.y *2 -1)

			if depthred == 1.0:
				return None
			else:
				view = self.uniforms['view']
				proj = self.uniforms['proj']
				a,b = proj[2][2], proj[3][2]
				depth = b/(depthred + a) * 0.5	# TODO get the true depth  (can't get why there is a strange factor ... opengl trick)
				#near, far = self.projection.limits  or settings.display['view_limits']
				#depth = 2 * near / (far + near - depthred * (far - near))
				#print('depth', depth, depthred)
				return fvec3(affineInverse(view) * fvec4(
							depth * x /proj[0][0],
							depth * y /proj[1][1],
							-depth,
							1))

		def ptfrom(self, point: QPoint, center: fvec3|vec3) -> fvec3:
			''' 3D point below the cursor in the plane orthogonal to the sight, with center as origin '''
			point = qpoint_to_vec(point)
			view = self.uniforms['view']
			proj = self.uniforms['proj']
			size = self.uniforms['size']
			x =  (point.x/size.x *2 -1)
			y = -(point.y/size.y *2 -1)
			depth = (view * fvec4(fvec3(center),1))[2]
			return fvec3(affineInverse(view) * fvec4(
						-depth * x /proj[0][0],
						-depth * y /proj[1][1],
						depth,
						1))

		def displayat(self, point: QPoint) -> (Display, int)|None:
			''' Return the display at the given screen position (widget relative).
			
				Every display can have a range of ident values in the ident map, they may correspond to subelements whose meaning only depends on the display. the subindex is returned with the display.
				
				If no object is at this exact location, None is returned.
			'''
			point = qpoint_to_vec(point)
			ident = self.ident.region((*point, *(point+1)))[0,0]
			if ident and 'ident' in self.scene.stacks:
				rdri = bisect(self.gl._steps, ident)
				if rdri == len(self.gl._steps):
					print('internal error: object ident points out of idents list')
					nprint(self.gl._steps)
				while rdri > 0 and self.gl._steps[rdri-1] == ident:	rdri -= 1
				if rdri > 0:	subi = ident - self.gl._steps[rdri-1] - 1
				else:			subi = ident - 1
				
				if rdri >= len(self.scene.stacks['ident']):
					print('wrong identification index', ident, self.scene.stacks['ident'][-1])
					nprint(self.gl._steps)
					return
				
				return (self.scene.stacks['ident'][rdri].display, int(subi))
				
		# move the view (in addition to what already exists in navigation)
		
		def look(self, position: fvec3=None):
			''' Make the scene navigation look at the position.
				This is changing the camera direction, center and distance.
			'''
			if not position:	position = self.scene.box().center
			self.navigation.look(position)
			self.update()
		
		def adjust(self, box:Box=None):
			''' Make the navigation camera large enough to get the given box in .
				This is changing the zoom level
			'''
			if not box:	box = self.scene.root.box
			if box.isempty():	return

			# get the most distant point to the focal axis
			invview = affineInverse(self.navigation.matrix())
			camera, look = fvec3(invview[3]), fvec3(invview[2])
			margin = 0.3  # margins around the zomed box, to make sure the displays fits in the view, sqrt(2)-1 should be necessary, but lower is generally better
			dist = length(noproject(box.center-camera, look)) + max(glm.abs(box.width))/2 * (1+margin)
			if not dist > 1e-6:	return

			# adjust navigation distance
			self.navigation.adjust(self.projection.adjust(dist))
			self.update()

		def center(self, center: fvec3=None):
			''' Relocate the navigation to the given position .
				This is translating the camera.
			'''
			if not center:	center = self.scene.box().center
			if not isfinite(center):	return

			self.navigation.center(center)
			self.update()


# view matrix providers

class Orbit:
	''' Navigation rotating on the 3 axis around a center.

		Object used as `View.navigation`
	'''
	position: fvec3
	distance: float
	orientation: fquat

	def __init__(self, position:fvec3=0, distance:float=1, orient:fvec3=fvec3(1,0,0)):
		self.center = fvec3(center)
		self.distance = float(distance)
		self.orient = fquat(orient)
		self.tool = navigation_tool

	def matrix(self) -> fmat4:
		mat = translate(fmat4(self.orient), -self.center)
		mat[3][2] -= self.distance
		return mat

	def adjust(self, distance: float):
		self.distance = distance

	def look(self, position: fvec3):
		dir = position - fvec3(affineInverse(self.matrix())[3])
		self.distance = length(dir)
		self.position = position
		self.sight(dir)
		
	def sight(self, direction: fvec3):
		if length2(direction, direction) <= 1e-6:
			return
		focal = self.orient * fvec3(0,0,1)
		self.orient = quat(direction, focal) * self.orient

	def center(self, position: fvec3):
		self.position = position

	def zoom(self, ratio: float):
		self.distance *= f

	def pan(self, offset: vec2):
		x,y,z = transpose(fmat3(self.orient))
		self.center += (fvec3(x) * -offset.x + fvec3(y) * offset.y) * self.distance/2

	def rotate(self, offset: vec3):
		# rotate from view euler angles
		self.orient = inverse(fquat(fvec3(-offset.y, -offset.dx, offset.dz) * pi)) * self.orient

class Turntable:
	''' Navigation rotating on yaw and pitch around a center 
	
		Object used as `View.navigation`
	'''
	position: fvec3
	distance: float
	pitch: float
	yaw: float
	
	def __init__(self, position:fvec3=0, distance:float=1, yaw:float=0, pitch:float=0):
		self.position = fvec3(position)
		self.yaw = yaw
		self.pitch = pitch
		self.distance = distance
	
	def matrix(self) -> fmat4:
		# build rotation from view euler angles
		rot = inverse(fquat(fvec3(pi/2-self.pitch, 0, -self.yaw)))
		mat = translate(fmat4(rot), -self.position)
		mat[3][2] -= self.distance
		return mat
	
	def adjust(self, distance: float):
		self.distance = distance
	
	def look(self, position: fvec3):
		dir = position - fvec3(affineInverse(self.matrix())[3])
		self.distance = length(dir)
		self.position = position
		self.sight(dir)
	
	def sight(self, direction: fvec3):
		if length2(direction) <= 1e-6:
			return
		self.yaw = atan2(direction.x, direction.y)
		self.pitch = -atan2(direction.z, length(direction.xy))
	
	def center(self, position: fvec3):
		self.position = position
	
	def zoom(self, ratio: float):
		self.distance *= ratio
	
	def pan(self, offset: fvec2):
		mat = transpose(fmat3(inverse(fquat(fvec3(pi/2-self.pitch, 0, -self.yaw)))))
		self.position += ( mat[0] * -offset.x + mat[1] * offset.y) * self.distance/2
	
	def rotate(self, offset: fvec3):
		if abs(self.pitch) > 3.2:	offset.x = -offset.x
		self.yaw += offset.x*pi
		self.pitch += offset.y*pi
		if self.pitch > pi:	self.pitch -= 2*pi
		if self.pitch < -pi: self.pitch += 2*pi

# projection matrix providers

class Perspective:
	''' Object used as `View.projection` 
	
		Attributes:
			fov (float):	field of view (rad), defaulting to `settings.display['field_of_view']`
	'''
	def __init__(self, fov=None):
		self.fov = fov or settings.display['field_of_view']
	
	def matrix(self, ratio, distance) -> fmat4:
		return perspective(self.fov/min(1, ratio), ratio, distance*1e-2, distance*1e4)
	
	def adjust(self, size):
		return size / tan(self.fov/2)

class Orthographic:
	''' Object used as `View.projection` 
	
		Attributes:
			size (float):  
			
				factor between the distance from camera to navigation center and the zone size to display
				defaulting to `tan(settings.display['field_of_view']/2)`
	'''
	def __init__(self, size=None):
		self.size = size or tan(settings.display['field_of_view']/2)

	def matrix(self, ratio, distance) -> fmat4:
		rf = 1e3 # relative far
		rn = 1e-2 # relative near
		return fmat4(min(1, 1/ratio)/(distance*self.size), 0, 0, 0,
					0,       min(1, ratio)/(distance*self.size), 0, 0,
					0,       0,          -2/(distance*(rf-rn)), 0,
					0,       0,          -(rf+rn)/(rf-rn), 1)
					
	def adjust(self, size):
		return size / self.size



'''
    minimal abstraction layer for the Qt wrapper
    
    it exposes the content of `QtCore`, `QtGui` and `QtWidgets`
    
    when imported this module proceeds as follow:
    - check if a supported Qt library is already loaded, then load it
    - else load the first available Qt library in the priority order
'''
import sys, importlib

alternatives = ['PyQt5', 'PySide6', 'PyQt6']

def use(lib):
    backend = importlib.import_module(lib)
    module = importlib.import_module(__name__+'.'+lib)
    globals().update(vars(module))

for lib in alternatives:
    if lib in sys.modules:
        use(lib)
        break
else:
    for lib in alternatives:
        try:
            use(lib)
        except ImportError as e:
            print(e)
            pass
        else:
            break
    else:
        raise ImportError('no Qt wrapper found: PySide6, PyQt5, PyQt6')
